from arro3.core import Scalar
from arro3.core.types import ArrayInput, ArrowStreamExportable

def max(input: ArrayInput | ArrowStreamExportable) -> Scalar:
    """
    Returns the max of values in the array.
    """

def min(input: ArrayInput | ArrowStreamExportable) -> Scalar:
    """
    Returns the min of values in the array.
    """

def sum(input: ArrayInput | ArrowStreamExportable) -> Scalar:
    """
    Returns the sum of values in the array.
    """
